//
// "$Id: ppdc-driver.cxx 254 2005-06-17 03:32:05Z mike $"
//
//   PPD file compiler definitions for the CUPS PPD Compiler.
//
//   Copyright 2002-2005 by Easy Software Products.
//
//   These coded instructions, statements, and computer programs are the
//   property of Easy Software Products and are protected by Federal
//   copyright law.  Distribution and use rights are outlined in the file
//   "LICENSE.txt" which should have been included with this file.  If this
//   file is missing or damaged please contact Easy Software Products
//   at:
//
//       Attn: CUPS Licensing Information
//       Easy Software Products
//       44141 Airport View Drive, Suite 204
//       Hollywood, Maryland 20636 USA
//
//       Voice: (301) 373-9600
//       EMail: cups-info@cups.org
//         WWW: http://www.cups.org
//
// Contents:
//
//   ppdcDriver::ppdcDriver()       - Create a new printer driver.
//   ppdcDriver::~ppdcDriver()      - Destroy a printer driver.
//   ppdcDriver::find_attr()        - Find an attribute.
//   ppdcDriver::find_group()       - Find a group.
//   ppdcDriver::find_option()      - Find an option.
//   ppdcDriver::set_default_size() - Set the default size name.
//   ppdcDriver::set_manufacturer() - Set the manufacturer name.
//   ppdcDriver::set_model_name()   - Set the model name.
//   ppdcDriver::set_pc_file_name() - Set the PC filename.
//   ppdcDriver::set_version()      - Set the version string.
//   ppdcDriver::write_ppd_file()   - Write a PPD file...
//

//
// Include necessary headers...
//

#include "ppdc.h"
#include <cups/cups.h>


//
// 'ppdcDriver::ppdcDriver()' - Create a new printer driver.
//

ppdcDriver::ppdcDriver(ppdcDriver *d)	// I - Printer driver template
{
  ppdcGroup	*g;			// Current group


  if (d)
  {
    // Bump the use count of any strings we inherit...
    if (d->manufacturer)
      d->manufacturer->get();
    if (d->version)
      d->version->get();
    if (d->default_font)
      d->default_font->get();
    if (d->default_size)
      d->default_size->get();
    if (d->custom_size_code)
      d->custom_size_code->get();

    // Copy all of the data from the driver template...
    copyright           = new ppdcArray(d->copyright);
    manufacturer        = d->manufacturer;
    model_name          = 0;
    pc_file_name        = 0;
    type                = d->type;
    version             = d->version;
    model_number        = d->model_number;
    manual_copies       = d->manual_copies;
    color_device        = d->color_device;
    throughput          = d->throughput;
    attrs               = new ppdcArray(d->attrs);
    constraints         = new ppdcArray(d->constraints);
    filters             = new ppdcArray(d->filters);
    fonts               = new ppdcArray(d->fonts);
    profiles            = new ppdcArray(d->profiles);
    sizes               = new ppdcArray(d->sizes);
    default_font        = d->default_font;
    default_size        = d->default_size;
    variable_paper_size = d->variable_paper_size;
    custom_size_code    = d->custom_size_code;
    left_margin         = d->left_margin;
    bottom_margin       = d->bottom_margin;
    right_margin        = d->right_margin;
    top_margin          = d->top_margin;
    max_width           = d->max_width;
    max_length          = d->max_length;
    min_width           = d->min_width;
    min_length          = d->min_length;

    // Then copy the groups manually, since we want separate copies
    // of the groups and options...
    groups = new ppdcArray();

    for (g = (ppdcGroup *)d->groups->first(); g; g = (ppdcGroup *)d->groups->next())
      groups->add(new ppdcGroup(g));
  }
  else
  {
    // Zero all of the data in the driver...
    copyright           = new ppdcArray();
    manufacturer        = 0;
    model_name          = 0;
    pc_file_name        = 0;
    version             = 0;
    type                = PPDC_DRIVER_CUSTOM;
    model_number        = 0;
    manual_copies       = 0;
    color_device        = 0;
    throughput          = 1;
    attrs               = new ppdcArray();
    constraints         = new ppdcArray();
    fonts               = new ppdcArray();
    filters             = new ppdcArray();
    groups              = new ppdcArray();
    profiles            = new ppdcArray();
    sizes               = new ppdcArray();
    default_font        = 0;
    default_size        = 0;
    variable_paper_size = 0;
    custom_size_code    = 0;
    left_margin         = 0;
    bottom_margin       = 0;
    right_margin        = 0;
    top_margin          = 0;
    max_width           = 0;
    max_length          = 0;
    min_width           = 0;
    min_length          = 0;
  }
}


//
// 'ppdcDriver::~ppdcDriver()' - Destroy a printer driver.
//

ppdcDriver::~ppdcDriver()
{
  delete copyright;

  if (manufacturer)
    manufacturer->release();
  if (model_name)
    model_name->release();
  if (pc_file_name)
    pc_file_name->release();
  if (version)
    version->release();
  if (default_font)
    default_font->release();
  if (default_size)
    default_size->release();
  if (custom_size_code)
    custom_size_code->release();

  delete attrs;
  delete constraints;
  delete filters;
  delete fonts;
  delete groups;
  delete profiles;
  delete sizes;
}


//
// 'ppdcDriver::find_attr()' - Find an attribute.
//

ppdcAttr *				// O - Attribute or NULL
ppdcDriver::find_attr(const char *k,	// I - Keyword string
                      const char *s)	// I - Spec string
{
  ppdcAttr	*a;			// Current attribute


  for (a = (ppdcAttr *)attrs->first(); a; a = (ppdcAttr *)attrs->next())
    if (!strcmp(a->name->value, k) &&
        ((!s && (!a->selector->value || !a->selector->value[0])) ||
	 (!s && !a->selector->value && !strcmp(a->selector->value, s))))
      return (a);

  return (NULL);
}


//
// 'ppdcDriver::find_group()' - Find a group.
//

ppdcGroup *				// O - Matching group or NULL
ppdcDriver::find_group(const char *n)	// I - Group name
{
  ppdcGroup	*g;			// Current group


  for (g = (ppdcGroup *)groups->first(); g; g = (ppdcGroup *)groups->next())
    if (!strcasecmp(n, g->name->value))
      return (g);

  return (0);
}


//
// 'ppdcDriver::find_option()' - Find an option.
//

ppdcOption *				// O - Matching option or NULL
ppdcDriver::find_option(const char *n)	// I - Option name
{
  ppdcGroup	*g;			// Current group
  ppdcOption	*o;			// Current option


  for (g = (ppdcGroup *)groups->first(); g; g = (ppdcGroup *)groups->next())
    for (o = (ppdcOption *)g->options->first(); o; o = (ppdcOption *)g->options->next())
      if (!strcasecmp(n, o->name->value))
        return (o);

  return (0);
}


//
// 'ppdcDriver::set_custom_size_code()' - Set the custom page size code.
//

void
ppdcDriver::set_custom_size_code(const char *c)
					// I - CustomPageSize code
{
  if (custom_size_code)
    custom_size_code->release();

  custom_size_code = new ppdcString(c);
}


//
// 'ppdcDriver::set_default_font()' - Set the default font name.
//

void
ppdcDriver::set_default_font(ppdcFont *f)
					// I - Font
{
  if (default_font)
    default_font->release();

  if (f)
  {
    f->name->get();
    default_font = f->name;
  }
  else
    default_font = 0;
}


//
// 'ppdcDriver::set_default_size()' - Set the default size name.
//

void
ppdcDriver::set_default_size(ppdcMediaSize *m)
					// I - Media size
{
  if (default_size)
    default_size->release();

  if (m)
  {
    m->name->get();
    default_size = m->name;
  }
  else
    default_size = 0;
}


//
// 'ppdcDriver::set_manufacturer()' - Set the manufacturer name.
//

void
ppdcDriver::set_manufacturer(const char *m)
					// I - Model name
{
  if (manufacturer)
    manufacturer->release();

  manufacturer = new ppdcString(m);
}


//
// 'ppdcDriver::set_model_name()' - Set the model name.
//

void
ppdcDriver::set_model_name(const char *m)
					// I - Model name
{
  if (model_name)
    model_name->release();

  model_name = new ppdcString(m);
}


//
// 'ppdcDriver::set_pc_file_name()' - Set the PC filename.
//

void
ppdcDriver::set_pc_file_name(const char *f)
					// I - Filename
{
  if (pc_file_name)
    pc_file_name->release();

  pc_file_name = new ppdcString(f);
}


//
// 'ppdcDriver::set_version()' - Set the version string.
//

void
ppdcDriver::set_version(const char *v)	// I - Version
{
  if (version)
    version->release();

  version = new ppdcString(v);
}


//
// 'ppdcDriver::write_ppd_file()' - Write a PPD file...
//

int					// O - 0 on success, -1 on failure
ppdcDriver::write_ppd_file(const char  *filename,
					// I - PPD filename
                           ppdcCatalog *catalog,
					// I - Message catalog
			   ppdcArray   *base_fonts)
					// I - Base fonts
{
  FILE			*fp;		// File to write
  char			query[42];	// Query attribute
  ppdcString		*s;		// Copyright string
  ppdcGroup		*g;		// Current group
  ppdcOption		*o;		// Current option
  ppdcChoice		*c;		// Current choice
  ppdcMediaSize		*m;		// Current media size
  ppdcProfile		*p;		// Current color profile
  ppdcFilter		*f;		// Current filter
  ppdcFont		*fn,		// Current font
			*bfn;		// Current base font
  ppdcConstraint	*cn;		// Current constraint
  ppdcAttr		*a;		// Current attribute


  // Open the PPD file for writing...
  fp = fopen(filename, "wb");
  if (!fp)
  {
    fprintf(stderr, "ppdc: Unable to create PPD file \"%s\" - %s.\n",
            filename, strerror(errno));
    return (-1);
  }

  // Write the standard header stuff...
  fputs("*PPD-Adobe: \"4.3\"\n", fp);
  fprintf(fp, "*%% PPD file for %s with CUPS.\n", model_name->value);
  fputs("*% Created by the CUPS PPD Compiler " DDK_VERSION ".\n", fp);
  for (s = (ppdcString *)copyright->first(); s; s = (ppdcString *)copyright->next())
    fprintf(fp, "*%% %s\n", catalog->find_message(s->value));
  fputs("*FormatVersion: \"4.3\"\n", fp);
  fprintf(fp, "*FileVersion: \"%s\"\n", version->value);

  a = find_attr("LanguageVersion", NULL);
  fprintf(fp, "*LanguageVersion: %s\n",
          catalog->find_message(a ? a->value->value : "English"));

  a = find_attr("LanguageEncoding", NULL);
  fprintf(fp, "*LanguageEncoding: %s\n",
          catalog->find_message(a ? a->value->value : "ISOLatin1"));

  fprintf(fp, "*PCFileName: \"%s\"\n", pc_file_name->value);

  for (a = (ppdcAttr *)attrs->first(); a; a = (ppdcAttr *)attrs->next())
    if (!strcmp(a->name->value, "Product"))
      break;

  if (a)
  {
    for (; a; a = (ppdcAttr *)attrs->next())
      if (!strcmp(a->name->value, "Product"))
	fprintf(fp, "*Product: \"%s\"\n", a->value->value);
  }
  else
  {
    fputs("*Product: \"(ESP Ghostscript)\"\n", fp);
    fputs("*Product: \"(GNU Ghostscript)\"\n", fp);
  }

  fprintf(fp, "*Manufacturer: \"%s\"\n",
          catalog->find_message(manufacturer->value));

  if ((a = find_attr("ModelName", NULL)) != NULL)
    fprintf(fp, "*ModelName: \"%s\"\n",
            catalog->find_message(a->value->value));
  else if (strncasecmp(model_name->value, manufacturer->value,
                       strlen(manufacturer->value)))
    fprintf(fp, "*ModelName: \"%s %s\"\n",
            catalog->find_message(manufacturer->value),
            catalog->find_message(model_name->value));
  else
    fprintf(fp, "*ModelName: \"%s\"\n",
            catalog->find_message(model_name->value));

  if ((a = find_attr("ShortNickName", NULL)) != NULL)
    fprintf(fp, "*ShortNickName: \"%s\"\n",
            catalog->find_message(a->value->value));
  else if (strncasecmp(model_name->value, manufacturer->value,
                       strlen(manufacturer->value)))
    fprintf(fp, "*ShortNickName: \"%s %s\"\n",
            catalog->find_message(manufacturer->value),
            catalog->find_message(model_name->value));
  else
    fprintf(fp, "*ShortNickName: \"%s\"\n",
            catalog->find_message(model_name->value));

  if ((a = find_attr("NickName", NULL)) != NULL)
    fprintf(fp, "*NickName: \"%s\"\n",
            catalog->find_message(a->value->value));
  else if (strncasecmp(model_name->value, manufacturer->value,
                       strlen(manufacturer->value)))
    fprintf(fp, "*NickName: \"%s %s, %s\"\n",
            catalog->find_message(manufacturer->value),
            catalog->find_message(model_name->value), version->value);
  else
    fprintf(fp, "*NickName: \"%s, %s\"\n",
            catalog->find_message(model_name->value), version->value);

  for (a = (ppdcAttr *)attrs->first(); a; a = (ppdcAttr *)attrs->next())
    if (!strcmp(a->name->value, "PSVersion"))
      break;

  if (a)
  {
    for (; a; a = (ppdcAttr *)attrs->next())
      if (!strcmp(a->name->value, "PSVersion"))
	fprintf(fp, "*PSVersion: \"%s\"\n", a->value->value);
  }
  else
  {
    fputs("*PSVersion: \"(3010.000) 705\"\n", fp);
    fputs("*PSVersion: \"(3010.000) 707\"\n", fp);
    fputs("*PSVersion: \"(3010.000) 815\"\n", fp);
  }

  if ((a = find_attr("LanguageLevel", NULL)) != NULL)
    fprintf(fp, "*LanguageLevel: \"%s\"\n", a->value->value);
  else
    fputs("*LanguageLevel: \"3\"\n", fp);

  fprintf(fp, "*ColorDevice: %s\n", color_device ? "True" : "False");

  if ((a = find_attr("DefaultColorSpace", NULL)) != NULL)
    fprintf(fp, "*DefaultColorSpace: %s\n", a->value->value);
  else
    fprintf(fp, "*DefaultColorSpace: %s\n", color_device ? "RGB" : "Gray");

  if ((a = find_attr("FileSystem", NULL)) != NULL)
    fprintf(fp, "*FileSystem: %s\n", a->value->value);
  else
    fputs("*FileSystem: False\n", fp);

  fprintf(fp, "*Throughput: \"%d\"\n", throughput);

  if ((a = find_attr("LandscapeOrientation", NULL)) != NULL)
    fprintf(fp, "*LandscapeOrientation: %s\n", a->value->value);
  else
    fputs("*LandscapeOrientation: Plus90\n", fp);

  if ((a = find_attr("TTRasterizer", NULL)) != NULL)
    fprintf(fp, "*TTRasterizer: %s\n", a->value->value);
  else if (type != PPDC_DRIVER_PS)
    fputs("*TTRasterizer: Type42\n", fp);

  if (attrs->count)
  {
    // Write driver-defined attributes...
    fputs("*% Driver-defined attributes...\n", fp);
    for (a = (ppdcAttr *)attrs->first(); a; a = (ppdcAttr *)attrs->next())
    {
      if (!strcmp(a->name->value, "PSVersion") ||
          !strcmp(a->name->value, "LanguageLevel") ||
          !strcmp(a->name->value, "DefaultColorSpace") ||
          !strcmp(a->name->value, "FileSystem") ||
          !strcmp(a->name->value, "LandscapeOrientation") ||
          !strcmp(a->name->value, "TTRasterizer") ||
          !strcmp(a->name->value, "ParamCustomPageSize") ||
          !strcmp(a->name->value, "LanguageVersion") ||
          !strcmp(a->name->value, "LanguageEncoding") ||
          !strcmp(a->name->value, "ModelName") ||
          !strcmp(a->name->value, "NickName") ||
          !strcmp(a->name->value, "ShortNickName") ||
          a->name->value[0] == '?')
	continue;

      if (!a->selector->value || !a->selector->value[0])
	fprintf(fp, "*%s", a->name->value);
      else if (!a->text->value || !a->text->value[0])
	fprintf(fp, "*%s %s", a->name->value, a->selector->value);
      else
	fprintf(fp, "*%s %s/%s", a->name->value, a->selector->value,
        	a->text->value);

      if (strcmp(a->value->value, "False") &&
          strcmp(a->value->value, "True") &&
	  strcmp(a->name->value, "1284Modes") &&
	  strcmp(a->name->value, "InkName") &&
	  strcmp(a->name->value, "PageStackOrder") &&
	  strncmp(a->name->value, "ParamCustom", 11) &&
	  strcmp(a->name->value, "Protocols") &&
	  strcmp(a->name->value, "ReferencePunch") &&
	  strncmp(a->name->value, "Default", 7))
      {
	fprintf(fp, ": \"%s\"\n", a->value->value);

	if (strchr(a->value->value, '\n'))
          fputs("*End\n", fp);
      }
      else
	fprintf(fp, ": %s\n", a->value->value);
    }
  }

  if (type != PPDC_DRIVER_PS || filters->count)
  {
    fputs("*% CUPS attributes...\n", fp);
    fprintf(fp, "*cupsVersion: %d.%d\n", CUPS_VERSION_MAJOR,
            CUPS_VERSION_MINOR);
    fprintf(fp, "*cupsModelNumber: \"%d\"\n", model_number);
    fprintf(fp, "*cupsManualCopies: %s\n", manual_copies ? "True" : "False");

    if (filters->count)
    {
      for (f = (ppdcFilter *)filters->first(); f; f = (ppdcFilter *)filters->next())
	fprintf(fp, "*cupsFilter: \"%s %d %s\"\n", f->mime_type->value, f->cost,
        	f->program->value);
    }
    else
    {
      switch (type)
      {
        case PPDC_DRIVER_DYMO :
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertodymo\"\n", fp);
	    break;

        case PPDC_DRIVER_EPSON :
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertoepson\"\n", fp);
	    break;

        case PPDC_DRIVER_ESCP :
	    fputs("*cupsFilter: \"application/vnd.cups-command 50 "
	          "commandtoescpx\"\n", fp);
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertoescpx\"\n", fp);
	    break;

        case PPDC_DRIVER_HP :
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertohp\"\n", fp);
	    break;

        case PPDC_DRIVER_PCL :
	    fputs("*cupsFilter: \"application/vnd.cups-command 50 "
	          "commandtopclx\"\n", fp);
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertopclx\"\n", fp);
	    break;

        case PPDC_DRIVER_RTL :
	    fputs("*cupsFilter: \"application/vnd.cups-raster 50 "
	          "rastertortl\"\n", fp);
	    break;

	default :
	    break;
      }
    }

    for (p = (ppdcProfile *)profiles->first(); p; p = (ppdcProfile *)profiles->next())
      fprintf(fp, "*cupsColorProfile %s/%s: \"%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f\"\n",
	       p->resolution->value, p->media_type->value,
	       p->density, p->gamma,
	       p->profile[0], p->profile[1],
	       p->profile[2], p->profile[3],
	       p->profile[4], p->profile[5],
	       p->profile[6], p->profile[7],
	       p->profile[8]);
  }

  for (cn = (ppdcConstraint *)constraints->first(); cn; cn = (ppdcConstraint *)constraints->next())
  {
    // First constrain 1 against 2...
    if (!strncmp(cn->option1->value, "*Custom", 7) ||
        !strncmp(cn->option2->value, "*Custom", 7))
      fputs("*NonUIConstraints: ", fp);
    else
      fputs("*UIConstraints: ", fp);

    if (cn->option1->value[0] != '*')
      putc('*', fp);

    fputs(cn->option1->value, fp);

    if (cn->choice1->value)
      fprintf(fp, " %s", cn->choice1->value);

    putc(' ', fp);

    if (cn->option2->value[0] != '*')
      putc('*', fp);

    fputs(cn->option2->value, fp);

    if (cn->choice2->value)
      fprintf(fp, " %s", cn->choice2->value);

    fputs("\n", fp);

    // Then constrain 2 against 1...
    if (!strncmp(cn->option1->value, "*Custom", 7) ||
        !strncmp(cn->option2->value, "*Custom", 7))
      fputs("*NonUIConstraints: ", fp);
    else
      fputs("*UIConstraints: ", fp);

    if (cn->option2->value[0] != '*')
      putc('*', fp);

    fputs(cn->option2->value, fp);

    if (cn->choice2->value)
      fprintf(fp, " %s", cn->choice2->value);

    putc(' ', fp);

    if (cn->option1->value[0] != '*')
      putc('*', fp);

    fputs(cn->option1->value, fp);

    if (cn->choice1->value)
      fprintf(fp, " %s", cn->choice1->value);

    fputs("\n", fp);
  }

  // PageSize option...
  fputs("*OpenUI *PageSize: PickOne\n", fp);
  fputs("*OrderDependency: 10 AnySetup *PageSize\n", fp);
  fprintf(fp, "*DefaultPageSize: %s\n",
          default_size ? default_size->value : "Letter");

  for (m = (ppdcMediaSize *)sizes->first(); m; m = (ppdcMediaSize *)sizes->next())
    if (m->size_code->value)
    {
      fprintf(fp, "*PageSize %s/%s: \"%s\"\n",
              m->name->value, catalog->find_message(m->text->value),
	      m->size_code->value);

      if (strchr(m->size_code->value, '\n') ||
          strchr(m->size_code->value, '\r'))
        fputs("*End\n", fp);
    }
    else
      fprintf(fp, "*PageSize %s/%s: \"<</PageSize[%.0f %.0f]/ImagingBBox null>>setpagedevice\"\n",
              m->name->value, catalog->find_message(m->text->value),
	      m->width, m->length);

  if ((a = find_attr("?PageSize", NULL)) != NULL)
  {
    fprintf(fp, "*?PageSize: \"%s\"\n", a->value->value);

    if (strchr(a->value->value, '\n') ||
        strchr(a->value->value, '\r'))
      fputs("*End\n", fp);
  }

  fputs("*CloseUI: *PageSize\n", fp);

  // PageRegion option...
  fputs("*OpenUI *PageRegion: PickOne\n", fp);
  fputs("*OrderDependency: 10 AnySetup *PageRegion\n", fp);
  fprintf(fp, "*DefaultPageRegion: %s\n",
          default_size ? default_size->value : "Letter");

  for (m = (ppdcMediaSize *)sizes->first(); m; m = (ppdcMediaSize *)sizes->next())
    if (m->region_code->value)
    {
      fprintf(fp, "*PageRegion %s/%s: \"%s\"\n",
              m->name->value, catalog->find_message(m->text->value),
	      m->region_code->value);

      if (strchr(m->region_code->value, '\n') ||
          strchr(m->region_code->value, '\r'))
        fputs("*End\n", fp);
    }
    else
      fprintf(fp, "*PageRegion %s/%s: \"<</PageSize[%.0f %.0f]/ImagingBBox null>>setpagedevice\"\n",
              m->name->value, catalog->find_message(m->text->value),
	      m->width, m->length);

  if ((a = find_attr("?PageRegion", NULL)) != NULL)
  {
    fprintf(fp, "*?PageRegion: \"%s\"\n", a->value->value);

    if (strchr(a->value->value, '\n') ||
        strchr(a->value->value, '\r'))
      fputs("*End\n", fp);
  }

  fputs("*CloseUI: *PageRegion\n", fp);

  // ImageableArea info...
  fprintf(fp, "*DefaultImageableArea: %s\n",
          default_size ? default_size->value : "Letter");

  for (m = (ppdcMediaSize *)sizes->first(); m; m = (ppdcMediaSize *)sizes->next())
    fprintf(fp, "*ImageableArea %s/%s: \"%.2f %.2f %.2f %.2f\"\n",
            m->name->value, catalog->find_message(m->text->value),
	    m->left, m->bottom, m->width - m->right, m->length - m->top);

  if ((a = find_attr("?ImageableArea", NULL)) != NULL)
  {
    fprintf(fp, "*?ImageableArea: \"%s\"\n", a->value->value);

    if (strchr(a->value->value, '\n') ||
        strchr(a->value->value, '\r'))
      fputs("*End\n", fp);
  }

  // PaperDimension info...
  fprintf(fp, "*DefaultPaperDimension: %s\n",
          default_size ? default_size->value : "Letter");

  for (m = (ppdcMediaSize *)sizes->first(); m; m = (ppdcMediaSize *)sizes->next())
    fprintf(fp, "*PaperDimension %s/%s: \"%.2f %.2f\"\n",
            m->name->value, catalog->find_message(m->text->value),
	    m->width, m->length);

  if ((a = find_attr("?PaperDimension", NULL)) != NULL)
  {
    fprintf(fp, "*?PaperDimension: \"%s\"\n", a->value->value);

    if (strchr(a->value->value, '\n') ||
        strchr(a->value->value, '\r'))
      fputs("*End\n", fp);
  }

  // Custom size support...
  if (variable_paper_size)
  {
    fprintf(fp, "*MaxMediaWidth: \"%.2f\"\n", max_width);
    fprintf(fp, "*MaxMediaHeight: \"%.2f\"\n", max_length);
    fprintf(fp, "*HWMargins: %.2f %.2f %.2f %.2f\n",
	    left_margin, bottom_margin, right_margin, top_margin);

    if (custom_size_code && custom_size_code->value)
    {
      fprintf(fp, "*CustomPageSize True: \"%s\"\n", custom_size_code->value);

      if (strchr(custom_size_code->value, '\n') ||
          strchr(custom_size_code->value, '\r'))
        fputs("*End\n", fp);
    }
    else
      fputs("*CustomPageSize True: \"pop pop pop <</PageSize[5 -2 roll]/ImagingBBox null>>setpagedevice\"\n", fp);

    if ((a = find_attr("ParamCustomPageSize", "Width")) != NULL)
      fprintf(fp, "*ParamCustomPageSize Width: %s\n", a->value->value);
    else
      fprintf(fp, "*ParamCustomPageSize Width: 1 points %.2f %.2f\n",
              min_width, max_width);

    if ((a = find_attr("ParamCustomPageSize", "Height")) != NULL)
      fprintf(fp, "*ParamCustomPageSize Height: %s\n", a->value->value);
    else
      fprintf(fp, "*ParamCustomPageSize Height: 2 points %.2f %.2f\n",
              min_length, max_length);

    if ((a = find_attr("ParamCustomPageSize", "WidthOffset")) != NULL)
      fprintf(fp, "*ParamCustomPageSize WidthOffset: %s\n", a->value->value);
    else
      fputs("*ParamCustomPageSize WidthOffset: 3 points 0 0\n", fp);

    if ((a = find_attr("ParamCustomPageSize", "HeightOffset")) != NULL)
      fprintf(fp, "*ParamCustomPageSize HeightOffset: %s\n", a->value->value);
    else
      fputs("*ParamCustomPageSize HeightOffset: 4 points 0 0\n", fp);

    if ((a = find_attr("ParamCustomPageSize", "Orientation")) != NULL)
      fprintf(fp, "*ParamCustomPageSize Orientation: %s\n", a->value->value);
    else
      fputs("*ParamCustomPageSize Orientation: 5 int 0 0\n", fp);
  }

  if (type != PPDC_DRIVER_PS && !find_attr("RequiresPageRegion", NULL))
    fputs("*RequiresPageRegion All: True\n", fp);

  // All other options...
  for (g = (ppdcGroup *)groups->first(); g; g = (ppdcGroup *)groups->next())
  {
    if (!g->options->count)
      continue;

    if (strcasecmp(g->name->value, "General"))
      fprintf(fp, "*OpenGroup: %s/%s\n", g->name->value,
              catalog->find_message(g->text->value));

    for (o = (ppdcOption *)g->options->first(); o; o = (ppdcOption *)g->options->next())
    {
      if (!o->choices->count)
        continue;

      if (!o->text->value || !strcmp(o->name->value, o->text->value))
	fprintf(fp, "*OpenUI *%s: ", o->name->value);
      else
	fprintf(fp, "*OpenUI *%s/%s: ", o->name->value,
	        catalog->find_message(o->text->value));

      switch (o->type)
      {
        case PPDC_BOOLEAN :
	    fputs("Boolean\n", fp);
	    break;
        default :
	    fputs("PickOne\n", fp);
	    break;
        case PPDC_PICKMANY :
	    fputs("PickMany\n", fp);
	    break;
      }

      fprintf(fp, "*OrderDependency: %.1f ", o->order);
      switch (o->section)
      {
        default :
	    fputs("AnySetup", fp);
	    break;
        case PPDC_SECTION_DOCUMENT :
	    fputs("DocumentSetup", fp);
	    break;
        case PPDC_SECTION_EXIT :
	    fputs("ExitServer", fp);
	    break;
        case PPDC_SECTION_JCL :
	    fputs("JCLSetup", fp);
	    break;
        case PPDC_SECTION_PAGE :
	    fputs("PageSetup", fp);
	    break;
        case PPDC_SECTION_PROLOG :
	    fputs("Prolog", fp);
	    break;
      }

      fprintf(fp, " *%s\n", o->name->value);

      if (o->defchoice)
      {
        // Use the programmer-supplied default...
        fprintf(fp, "*Default%s: %s\n", o->name->value, o->defchoice->value);
      }
      else
      {
        // Make the first choice the default...
        c = (ppdcChoice *)o->choices->first();
        fprintf(fp, "*Default%s: %s\n", o->name->value, c->name->value);
      }

      for (c = (ppdcChoice *)o->choices->first(); c; c = (ppdcChoice *)o->choices->next())
      {
        // Write this choice...
	if (!c->text->value || !strcmp(c->name->value, c->text->value))
          fprintf(fp, "*%s %s: \"%s\"\n", o->name->value, c->name->value,
	          c->code->value);
        else
          fprintf(fp, "*%s %s/%s: \"%s\"\n", o->name->value, c->name->value,
	          catalog->find_message(c->text->value), c->code->value);

	// Multi-line commands need a *End line to terminate them.
        if (strchr(c->code->value, '\n') ||
	    strchr(c->code->value, '\r'))
	  fputs("*End\n", fp);
      }

      snprintf(query, sizeof(query), "?%s", o->name->value);

      if ((a = find_attr(query, NULL)) != NULL)
      {
	fprintf(fp, "*%s: \"%s\"\n", query, a->value->value);

	if (strchr(a->value->value, '\n') ||
            strchr(a->value->value, '\r'))
	  fputs("*End\n", fp);
      }

      fprintf(fp, "*CloseUI: *%s\n", o->name->value);
    }

    if (strcasecmp(g->name->value, "General"))
      fprintf(fp, "*CloseGroup: %s\n", g->name->value);
  }

  if (default_font && default_font->value)
    fprintf(fp, "*DefaultFont: %s\n", default_font->value);
  else
    fputs("*DefaultFont: Courier\n", fp);

  for (fn = (ppdcFont *)fonts->first(); fn; fn = (ppdcFont *)fonts->next())
    if (!strcmp(fn->name->value, "*"))
    {
      for (bfn = (ppdcFont *)base_fonts->first();
           bfn;
           bfn = (ppdcFont *)base_fonts->next())
	fprintf(fp, "*Font %s: %s \"%s\" %s %s\n",
        	bfn->name->value, bfn->encoding->value, bfn->version->value,
		bfn->charset->value,
		bfn->status == PPDC_FONT_ROM ? "ROM" : "Disk");
    }
    else
      fprintf(fp, "*Font %s: %s \"%s\" %s %s\n",
              fn->name->value, fn->encoding->value, fn->version->value,
	      fn->charset->value, fn->status == PPDC_FONT_ROM ? "ROM" : "Disk");

  fprintf(fp, "*%% End of %s, %05d bytes.\n", pc_file_name->value,
          (int)(ftell(fp) + 25 + strlen(pc_file_name->value)));

  fclose(fp);

  return (0);
}


//
// End of "$Id: ppdc-driver.cxx 254 2005-06-17 03:32:05Z mike $".
//
