//
// "$Id: ppdc.cxx 255 2005-06-23 17:51:23Z mike $"
//
//   PPD file compiler main entry for the CUPS PPD Compiler.
//
//   Copyright 2002-2005 by Easy Software Products.
//
//   These coded instructions, statements, and computer programs are the
//   property of Easy Software Products and are protected by Federal
//   copyright law.  Distribution and use rights are outlined in the file
//   "LICENSE.txt" which should have been included with this file.  If this
//   file is missing or damaged please contact Easy Software Products
//   at:
//
//       Attn: CUPS Licensing Information
//       Easy Software Products
//       44141 Airport View Drive, Suite 204
//       Hollywood, Maryland 20636 USA
//
//       Voice: (301) 373-9600
//       EMail: cups-info@cups.org
//         WWW: http://www.cups.org
//
// Contents:
//
//   main()  - Main entry for the PPD compiler.
//   usage() - Show usage and exit.
//

//
// Include necessary headers...
//

#include "ppdc.h"
#include <sys/stat.h>
#include <sys/types.h>


//
// Local functions...
//

static void	usage(void);


//
// 'main()' - Main entry for the PPD compiler.
//

int					// O - Exit status
main(int  argc,				// I - Number of command-line arguments
     char *argv[])			// I - Command-line arguments
{
  int		i, j;			// Looping vars
  ppdcCatalog	*catalog;		// Message catalog
  const char	*outdir;		// Output directory
  ppdcSource	*src;			// PPD source file data
  ppdcDriver	*d;			// Current driver
  char		*opt,			// Current option
		pcfilename[1024],	// Lowercase pcfilename
		filename[1024];		// PPD filename
  int		verbose;		// Verbosity


  // Scan the command-line...
  catalog = new ppdcCatalog();
  outdir  = "ppd";
  src     = 0;
  verbose = 0;

  for (i = 1; i < argc; i ++)
    if (argv[i][0] == '-')
    {
      for (opt = argv[i] + 1; *opt; opt ++)
        switch (*opt)
	{
	  case 'c' :			// Message catalog...
	      i ++;
              if (i >= argc)
                usage();

              if (verbose > 1)
	        printf("ppdc: Loading messages from \"%s\"...\n", argv[i]);

              if (catalog->load_messages(argv[i]))
	      {
        	fprintf(stderr, "ppdc: Unable to load localization file \"%s\" - %s\n",
	        	argv[i], strerror(errno));
                return (1);
	      }
	      break;

          case 'd' :			// Output directory...
	      i ++;
	      if (i >= argc)
        	usage();

              if (verbose > 1)
	        printf("ppdc: Writing PPD files to directory \"%s\"...\n",
		       argv[i]);

	      outdir = argv[i];
	      break;

          case 'l' :			// Language...
	      i ++;
	      if (i >= argc)
        	usage();

	      snprintf(filename, sizeof(filename), PPDC_PO_DIR "/%s.po",
	               argv[i]);

              if (verbose > 1)
	        printf("ppdc: Loading messages from \"%s\"...\n", filename);

	      if (catalog->load_messages(filename))
	      {
        	fprintf(stderr, "ppdc: Unable to find localization for \"%s\" - %s\n",
	        	argv[i], strerror(errno));
                return (1);
	      }
	      break;

          case 'I' :			// Include directory...
	      i ++;
	      if (i >= argc)
        	usage();

              if (verbose > 1)
	        printf("ppdc: Adding include directory \"%s\"...\n", argv[i]);

	      ppdcSource::add_include(argv[i]);
	      break;

          case 'v' :			// Be verbose...
	      verbose ++;
	      break;

	  default :			// Unknown
	      usage();
	      break;
	}
    }
    else
    {
      // Open and load the driver info file...
      if (verbose > 1)
        printf("ppdc: Loading driver information file \"%s\"...\n", argv[i]);

      src = new ppdcSource(argv[i]);

      // Create the output directory...
      if (mkdir(outdir, 0777))
      {
        if (errno != EEXIST)
	{
	  fprintf(stderr, "ppdc: Unable to create output directory %s: %s\n",
	          outdir, strerror(errno));
          return (1);
	}
      }

      // Write PPD files...
      for (d = (ppdcDriver *)src->drivers->first(); d; d = (ppdcDriver *)src->drivers->next())
      {
        // Write the PPD file for this driver...
	for (j = 0; d->pc_file_name->value[j]; j ++)
	  pcfilename[j] = tolower(d->pc_file_name->value[j]);

	pcfilename[j] = '\0';

        snprintf(filename, sizeof(filename), "%s/%s", outdir, pcfilename);

	if (verbose)
	  printf("ppdc: Writing %s...\n", filename);

	if (d->write_ppd_file(filename, catalog, src->base_fonts))
	  return (1);
      }

      // Delete the printer driver information...
      delete src;
    }

  delete catalog;

  // If no drivers have been loaded, display the program usage message.
  if (!src)
    usage();

  // Return with no errors.
  return (0);
}


//
// 'usage()' - Show usage and exit.
//

static void
usage(void)
{
  puts("Usage: ppdc [options] filename.drv [ ... filenameN.drv ]");
  puts("Options:");
  puts("  -I include-dir    Add include directory to search path.");
  puts("  -c catalog.po     Load the specified message catalog.");
  puts("  -d output-dir     Specify the output directory.");
  puts("  -l language       Specify the output language (locale).");
  puts("  -v                Be verbose (more v's for more verbosity).");

  exit(1);
}


//
// End of "$Id: ppdc.cxx 255 2005-06-23 17:51:23Z mike $".
//
